@@ -101,7 +101,7 @@ defmodule ExUnit.Callbacks do
         setup context do
           IO.puts("Setting up: #{context.test}")
 
-          # We can simply return :ok when we don't want add any extra metadata
+          # We can simply return :ok when we don't want to add any extra metadata
           :ok
         end
 
@@ -134,7 +134,7 @@ defmodule ExUnit.Callbacks do
 
         defp step1(_context), do: [step_one: true]
         defp step2(_context), do: {:ok, step_two: true} # return values with shape of {:ok, keyword() | map()} allowed
-        defp step3(_context), do: :ok  # Context not modified
+        defp step3(_context), do: :ok # Context not modified
 
         test "context was modified", context do
           assert context[:step_one] == true
@@ -167,6 +167,7 @@ defmodule ExUnit.Callbacks do
       @ex_unit_describe nil
       @ex_unit_setup []
       @ex_unit_setup_all []
+      @ex_unit_used_describes %{}
 
       @before_compile unquote(__MODULE__)
       import unquote(__MODULE__)
@@ -175,7 +176,8 @@ defmodule ExUnit.Callbacks do
 
   @doc false
   defmacro __before_compile__(env) do
-    [compile_callbacks(env, :setup), compile_callbacks(env, :setup_all)]
+    used_describes = Module.get_attribute(env.module, :ex_unit_used_describes)
+    [compile_setup(env, :setup, used_describes), compile_setup(env, :setup_all, %{})]
   end
 
   @doc """
@@ -208,8 +210,7 @@ defmodule ExUnit.Callbacks do
       do_setup(quote(do: _), block)
     else
       quote do
-        @ex_unit_setup ExUnit.Callbacks.__callback__(unquote(block), @ex_unit_describe) ++
-                         @ex_unit_setup
+        ExUnit.Callbacks.__setup__(__MODULE__, unquote(block))
       end
     end
   end
@@ -236,12 +237,31 @@ defmodule ExUnit.Callbacks do
 
   defp do_setup(context, block) do
     quote bind_quoted: [context: escape(context), block: escape(block)] do
-      name = :"__ex_unit_setup_#{length(@ex_unit_setup)}"
+      name = ExUnit.Callbacks.__setup__(__MODULE__)
       defp unquote(name)(unquote(context)), unquote(block)
-      @ex_unit_setup [{name, @ex_unit_describe} | @ex_unit_setup]
     end
   end
 
+  @doc false
+  def __setup__(module, callbacks) do
+    setup = Module.get_attribute(module, :ex_unit_setup)
+    Module.put_attribute(module, :ex_unit_setup, Enum.reverse(callbacks(callbacks), setup))
+  end
+
+  @doc false
+  def __setup__(module) do
+    setup = Module.get_attribute(module, :ex_unit_setup)
+
+    name =
+      case Module.get_attribute(module, :ex_unit_describe) do
+        {_line, _message, counter} -> :"__ex_unit_setup_#{counter}_#{length(setup)}"
+        nil -> :"__ex_unit_setup_#{length(setup)}"
+      end
+
+    Module.put_attribute(module, :ex_unit_setup, [name | setup])
+    name
+  end
+
   @doc """
   Defines a callback to be run before all tests in a case.
 
@@ -254,6 +274,15 @@ defmodule ExUnit.Callbacks do
   `setup_all/1` callbacks are executed in a separate process than tests.
   All `setup_all/1` callbacks are executed in order in the same process.
 
+  ## On-Exit Handlers
+
+  On-exit handlers that you register inside `setup_all/1` callbacks
+  are executed at once after all tests in the module have been run.
+  They are all executed *in the same process*, which is a separate
+  process dedicated to running these handlers. These handlers are
+  executed in the reverse order of their respective `setup_all/1`
+  callbacks.
+
   ## Examples
 
       # One-arity function name
@@ -277,18 +306,29 @@ defmodule ExUnit.Callbacks do
         # ...
       end
 
+  ### Handlers
+
+  You can define "global" on-exit handlers in `setup_all/1` callbacks:
+
+      setup_all do
+        Database.create_table_for(__MODULE__)
+
+        on_exit(fn ->
+          Database.drop_table_for(__MODULE__)
+        end)
+
+        :ok
+      end
+
+  The handler in the example above will be executed only once, after
+  running all tests in the module.
   """
   defmacro setup_all(block) do
     if Keyword.keyword?(block) do
       do_setup_all(quote(do: _), block)
     else
       quote do
-        @ex_unit_describe &&
-          raise "cannot invoke setup_all/1 inside describe as setup_all/1 " <>
-                  "always applies to all tests in a module"
-
-        @ex_unit_setup_all ExUnit.Callbacks.__callback__(unquote(block), nil) ++
-                             @ex_unit_setup_all
+        ExUnit.Callbacks.__setup_all__(__MODULE__, unquote(block))
       end
     end
   end
@@ -312,32 +352,76 @@ defmodule ExUnit.Callbacks do
 
   defp do_setup_all(context, block) do
     quote bind_quoted: [context: escape(context), block: escape(block)] do
-      @ex_unit_describe && raise "cannot invoke setup_all/2 inside describe"
-      name = :"__ex_unit_setup_all_#{length(@ex_unit_setup_all)}"
+      name = ExUnit.Callbacks.__setup_all__(__MODULE__)
       defp unquote(name)(unquote(context)), unquote(block)
-      @ex_unit_setup_all [{name, nil} | @ex_unit_setup_all]
+    end
+  end
+
+  @doc false
+  def __setup_all__(module, callbacks) do
+    no_describe!(module)
+    setup_all = Module.get_attribute(module, :ex_unit_setup_all)
+
+    Module.put_attribute(
+      module,
+      :ex_unit_setup_all,
+      Enum.reverse(callbacks(callbacks), setup_all)
+    )
+  end
+
+  @doc false
+  def __setup_all__(module) do
+    no_describe!(module)
+    setup_all = Module.get_attribute(module, :ex_unit_setup_all)
+    name = :"__ex_unit_setup_all_#{length(setup_all)}"
+    Module.put_attribute(module, :ex_unit_setup_all, [name | setup_all])
+    name
+  end
+
+  defp no_describe!(module) do
+    if Module.get_attribute(module, :ex_unit_describe) do
+      raise "cannot invoke setup_all/1-2 inside describe as setup_all " <>
+              "always applies to all tests in a module"
+    end
+  end
+
+  defp callbacks(callbacks) do
+    for k <- List.wrap(callbacks) do
+      if not is_atom(k) do
+        raise ArgumentError,
+              "setup/setup_all expect a callback name as an atom or " <>
+                "a list of callback names, got: #{inspect(k)}"
+      end
+
+      k
     end
   end
 
   @doc """
-  Defines a callback that runs once the test exits.
+  Registers a callback that runs once the test exits.
 
   `callback` is a function that receives no arguments and
-  runs in a separate process than the caller.
+  runs in a separate process than the caller. Its return
+  value is irrelevant and is discarded.
 
-  `on_exit/2` is usually called from `setup` and `setup_all`
+  `on_exit/2` is usually called from `setup/1` and `setup_all/1`
   callbacks, often to undo the action performed during the setup.
-  However, `on_exit/2` may also be called dynamically, where a
-  reference can be used to guarantee the callback will be invoked
-  only once.
 
-  `on_exit/2` gets executed in a blocking fashion after a test
-  exits and **before** running the next test. This means that no
-  other test from the same test case will be running while the
-  `on_exit/2` callback for a previous test is running.
+  However, `on_exit/2` may also be called dynamically. An "ID" (the
+  `name_or_ref` argument) can be used to guarantee that the callback
+  will be invoked only once. ExUnit uses this term to identify an
+  `on_exit/2` handler: if you want to override a previous handler, for
+  example, use the same `name_or_ref` across multiple `on_exit/2`
+  calls.
 
-  `on_exit/2` is executed in a different process than the test
-  process.
+  If `on_exit/2` is called inside `setup/1` or inside a test, it's
+  executed in a blocking fashion after the test exits and *before
+  running the next test*. This means that no other test from the same
+  test case will be running while the `on_exit/2` callback for a
+  previous test is running. `on_exit/2` is executed in a different
+  process than the test process. On the other hand, if `on_exit/2` is
+  called inside a `setup_all/1` callback then `callback` is executed
+  after running *all tests* (see `setup_all/1` for more information).
 
   ## Examples
 
@@ -346,6 +430,23 @@ defmodule ExUnit.Callbacks do
         on_exit(fn -> File.rm!("fixture.json") end)
       end
 
+  You can use the same `name_or_ref` across multiple `on_exit/2` calls
+  to "override" the registered handler:
+
+      setup do
+        on_exit(:drop_table, fn ->
+          Database.drop_table()
+        end)
+      end
+
+      test "a test that shouldn't drop the table" do
+        on_exit(:drop_table, fn -> :ok end)
+      end
+
+  Relying too much on overriding callbacks like this can lead to test
+  cases that are hard to understand and with too many layers of
+  indirection. However, it can be useful in some cases or for library
+  authors, for example.
   """
   @spec on_exit(term, (() -> term)) :: :ok
   def on_exit(name_or_ref \\ make_ref(), callback) when is_function(callback, 0) do
@@ -388,6 +489,10 @@ defmodule ExUnit.Callbacks do
   test, as simply shutting down the process would cause it to be restarted
   according to its `:restart` value.
 
+  The started process is not linked to the test process and a crash will
+  not necessarily fail the test. To start and link a process to guarantee
+  that any crash would also fail the test use `start_link_supervised!/2`.
+
   This function returns `{:ok, pid}` in case of success, otherwise it
   returns `{:error, reason}`.
   """
@@ -443,6 +548,21 @@ defmodule ExUnit.Callbacks do
 
   defp start_supervised_error(reason), do: Exception.format_exit({:start_spec, reason})
 
+  @doc """
+  Same as `start_supervised!/2` but links the started process to the test process.
+
+  This means that if the process that was started crashes that crash is propagated to
+  the test process, failing the test and printing the cause of the crash.
+  """
+  @doc since: "1.14.0"
+  @spec start_link_supervised!(Supervisor.child_spec() | module | {module, term}, keyword) ::
+          pid
+  def start_link_supervised!(child_spec_or_module, opts \\ []) do
+    pid = start_supervised!(child_spec_or_module, opts)
+    Process.link(pid)
+    pid
+  end
+
   @doc """
   Stops a child process started via `start_supervised/2`.
 
@@ -463,15 +583,32 @@ defmodule ExUnit.Callbacks do
         {:error, :not_found}
 
       {:ok, sup} ->
-        with :ok <- Supervisor.terminate_child(sup, id),
-             :ok <- Supervisor.delete_child(sup, id),
-             do: :ok
+        pid = pid_for_child(sup, id)
+
+        if pid do
+          Process.unlink(pid)
+        end
+
+        with :ok <- Supervisor.terminate_child(sup, id) do
+          # If the terminated child was temporary, delete_child returns {:error, :not_found}.
+          # Since the child was successfully terminated, we treat this result as a success.
+          true = Supervisor.delete_child(sup, id) in [:ok, {:error, :not_found}]
+          :ok
+        end
 
       :error ->
         raise ArgumentError, "stop_supervised/1 can only be invoked from the test process"
     end
   end
 
+  defp pid_for_child(sup, id) do
+    children = Supervisor.which_children(sup)
+
+    with {_id, pid, _type, _modules} <- List.keyfind(children, id, 0) do
+      pid
+    end
+  end
+
   @doc """
   Same as `stop_supervised/1` but raises if it cannot be stopped.
   """
@@ -506,31 +643,31 @@ defmodule ExUnit.Callbacks do
   end
 
   @doc false
-  def __merge__(mod, context, value) do
-    merge(mod, context, value, value)
-  end
-
-  defp merge(_mod, context, :ok, _original_value) do
+  def __merge__(_mod, context, :ok) do
     context
   end
 
-  defp merge(mod, context, {:ok, value}, original_value) do
-    merge(mod, context, value, original_value)
+  def __merge__(mod, context, {:ok, value}) do
+    unwrapped_merge(mod, context, value, {:ok, value})
   end
 
-  defp merge(mod, _context, %_{}, original_value) do
+  def __merge__(mod, context, value) do
+    unwrapped_merge(mod, context, value, value)
+  end
+
+  defp unwrapped_merge(mod, _context, %_{}, original_value) do
     raise_merge_failed!(mod, original_value)
   end
 
-  defp merge(mod, context, data, original_value) when is_list(data) do
-    merge(mod, context, Map.new(data), original_value)
+  defp unwrapped_merge(mod, context, data, _original_value) when is_list(data) do
+    context_merge(mod, context, Map.new(data))
   end
 
-  defp merge(mod, context, data, _original_value) when is_map(data) do
+  defp unwrapped_merge(mod, context, data, _original_value) when is_map(data) do
     context_merge(mod, context, data)
   end
 
-  defp merge(mod, _, _return_value, original_value) do
+  defp unwrapped_merge(mod, _, _return_value, original_value) do
     raise_merge_failed!(mod, original_value)
   end
 
@@ -558,44 +695,112 @@ defmodule ExUnit.Callbacks do
     Macro.escape(contents, unquote: true)
   end
 
-  defp compile_callbacks(env, kind) do
-    callbacks = Module.get_attribute(env.module, :"ex_unit_#{kind}") |> Enum.reverse()
+  @doc false
+  def __describe__(module, line, message, fun) do
+    if Module.get_attribute(module, :ex_unit_describe) do
+      raise "cannot call \"describe\" inside another \"describe\". See the documentation " <>
+              "for ExUnit.Case.describe/2 on named setups and how to handle hierarchies"
+    end
 
-    acc =
-      case callbacks do
-        [] ->
-          quote(do: context)
+    used_describes = Module.get_attribute(module, :ex_unit_used_describes)
 
-        [h | t] ->
-          Enum.reduce(t, compile_merge(h), fn callback_describe, acc ->
-            quote do
-              context = unquote(acc)
-              unquote(compile_merge(callback_describe))
-            end
-          end)
-      end
+    cond do
+      not is_binary(message) ->
+        raise ArgumentError, "describe name must be a string, got: #{inspect(message)}"
 
-    quote do
-      def __ex_unit__(unquote(kind), context) do
-        describe = Map.get(context, :describe, nil)
-        unquote(acc)
+      is_map_key(used_describes, message) ->
+        raise ExUnit.DuplicateDescribeError,
+              "describe #{inspect(message)} is already defined in #{inspect(module)}"
+
+      true ->
+        :ok
+    end
+
+    if Module.get_attribute(module, :describetag) != [] do
+      raise "@describetag must be set inside describe/2 blocks"
+    end
+
+    setup = Module.get_attribute(module, :ex_unit_setup)
+    Module.put_attribute(module, :ex_unit_setup, [])
+    Module.put_attribute(module, :ex_unit_describe, {line, message, map_size(used_describes)})
+
+    try do
+      fun.(message, used_describes)
+    after
+      Module.put_attribute(module, :ex_unit_describe, nil)
+      Module.put_attribute(module, :ex_unit_setup, setup)
+      Module.delete_attribute(module, :describetag)
+
+      for attribute <- Module.get_attribute(module, :ex_unit_registered_describe_attributes) do
+        Module.delete_attribute(module, attribute)
       end
     end
   end
 
-  defp compile_merge({callback, nil}) do
-    quote do
-      unquote(__MODULE__).__merge__(__MODULE__, context, unquote(callback)(context))
-    end
+  @doc false
+  def __describe__(module, message, used_describes) do
+    {name, body} =
+      case Module.get_attribute(module, :ex_unit_setup) do
+        [] -> {nil, nil}
+        callbacks -> {:"__ex_unit_describe_#{map_size(used_describes)}", compile_setup(callbacks)}
+      end
+
+    used_describes = Map.put(used_describes, message, name)
+    Module.put_attribute(module, :ex_unit_used_describes, used_describes)
+    {name, body}
   end
 
-  defp compile_merge({callback, {_line, describe}}) do
-    quote do
-      if unquote(describe) == describe do
-        unquote(compile_merge({callback, nil}))
+  defp compile_setup(env, kind, describes) do
+    calls =
+      env.module
+      |> Module.get_attribute(:"ex_unit_#{kind}")
+      |> compile_setup()
+
+    describe_clauses =
+      for {describe, callback} <- describes,
+          callback != nil,
+          clause <- quote(do: (unquote(describe) -> unquote(callback)(var!(context)))),
+          do: clause
+
+    body =
+      if describe_clauses == [] do
+        calls
       else
-        context
+        describe_clauses =
+          describe_clauses ++
+            quote do
+              _ -> var!(context)
+            end
+
+        quote do
+          var!(context) = unquote(calls)
+          case Map.get(var!(context), :describe, nil), do: unquote(describe_clauses)
+        end
       end
+
+    quote do
+      def __ex_unit__(unquote(kind), var!(context)), do: unquote(body)
+    end
+  end
+
+  defp compile_setup([]) do
+    quote(do: var!(context))
+  end
+
+  defp compile_setup(callbacks) do
+    [h | t] = Enum.reverse(callbacks)
+
+    Enum.reduce(t, compile_setup_call(h), fn callback, acc ->
+      quote do
+        var!(context) = unquote(acc)
+        unquote(compile_setup_call(callback))
+      end
+    end)
+  end
+
+  defp compile_setup_call(callback) do
+    quote do
+      unquote(__MODULE__).__merge__(__MODULE__, var!(context), unquote(callback)(var!(context)))
     end
   end
 end
